@@ -26,21 +26,23 @@
 #include <config.h>
 #include "eel-gdk-pixbuf-extensions.h"
 
+#include "eel-art-gtk-extensions.h"
+#include "eel-debug-drawing.h"
+#include "eel-debug.h"
 #include "eel-gdk-extensions.h"
 #include "eel-glib-extensions.h"
-#include "eel-string.h"
-#include "eel-art-gtk-extensions.h"
 #include "eel-lib-self-check-functions.h"
-#include "eel-debug-drawing.h"
+#include "eel-string.h"
 #include <gdk-pixbuf/gdk-pixbuf-loader.h>
-#include <gdk/gdkx.h>
 #include <gdk/gdkprivate.h>
+#include <gdk/gdkx.h>
 #include <libgnomevfs/gnome-vfs-async-ops.h>
 #include <libgnomevfs/gnome-vfs-ops.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
 #include <math.h>
+#include <stdlib.h>
+#include <stdio.h>
 #include <pango/pangoft2.h>
-#include <png.h>
 
 #define LOAD_BUFFER_SIZE 65536
 
@@ -79,7 +81,7 @@ static void load_done            (EelPixbufLoadHandle *handle,
 void
 eel_gdk_pixbuf_list_ref (GList *pixbuf_list)
 {
-	g_list_foreach (pixbuf_list, (GFunc) gdk_pixbuf_ref, NULL);
+	g_list_foreach (pixbuf_list, (GFunc) g_object_ref, NULL);
 }
 
 /**
@@ -91,7 +93,7 @@ eel_gdk_pixbuf_list_ref (GList *pixbuf_list)
 void
 eel_gdk_pixbuf_list_free (GList *pixbuf_list)
 {
-	eel_g_list_free_deep_custom (pixbuf_list, (GFunc) gdk_pixbuf_unref, NULL);
+	eel_g_list_free_deep_custom (pixbuf_list, (GFunc) g_object_unref, NULL);
 }
 
 GdkPixbuf *
@@ -135,6 +137,7 @@ eel_gdk_pixbuf_load (const char *uri)
 	}
 
 	if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF) {
+		gdk_pixbuf_loader_close (loader, NULL);
 		g_object_unref (loader);
 		gnome_vfs_close (handle);
 		return NULL;
@@ -219,13 +222,14 @@ file_read_callback (GnomeVFSAsyncHandle *vfs_handle,
 					      bytes_read,
 					      NULL)) {
 			result = GNOME_VFS_ERROR_WRONG_FORMAT;
+		} else {
+			gnome_vfs_async_read (handle->vfs_handle,
+					      handle->buffer,
+					      sizeof (handle->buffer),
+					      file_read_callback,
+					      handle);
+			return;
 		}
-		gnome_vfs_async_read (handle->vfs_handle,
-				      handle->buffer,
-				      sizeof (handle->buffer),
-				      file_read_callback,
-				      handle);
-		return;
 	}
 
 	load_done (handle, result, result == GNOME_VFS_OK || result == GNOME_VFS_ERROR_EOF);
@@ -274,6 +278,9 @@ eel_cancel_gdk_pixbuf_load (EelPixbufLoadHandle *handle)
 	if (handle == NULL) {
 		return;
 	}
+	if (handle->loader != NULL) {
+		gdk_pixbuf_loader_close (handle->loader, NULL);
+	}
 	if (handle->vfs_handle != NULL) {
 		gnome_vfs_async_cancel (handle->vfs_handle);
 	}
@@ -285,42 +292,58 @@ guint32
 eel_gdk_pixbuf_average_value (GdkPixbuf *pixbuf)
 {
 	guint64 a_total, r_total, g_total, b_total;
-	int row, column;
-	gboolean has_alpha;
-	int width, height;
+	guint row, column;
 	int row_stride;
 	const guchar *pixels, *p;
 	int r, g, b, a;
-	int dividend;
+	guint dividend;
+	guint width, height;
 
-	has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
 	width = gdk_pixbuf_get_width (pixbuf);
 	height = gdk_pixbuf_get_height (pixbuf);
 	row_stride = gdk_pixbuf_get_rowstride (pixbuf);
 	pixels = gdk_pixbuf_get_pixels (pixbuf);
-	
+
 	/* iterate through the pixbuf, counting up each component */
 	a_total = 0;
 	r_total = 0;
 	g_total = 0;
 	b_total = 0;
-	for (row = 0; row < height; row++) {
-		p = pixels + (row * row_stride);
-		for (column = 0; column < width; column++) {
-			r = *p++;
-			g = *p++;
-			b = *p++;
-			a = has_alpha ? *p++ : 0xFF;
 
-			a_total += a;
-			r_total += r * a;
-			g_total += g * a;
-			b_total += b * a;
+	if (gdk_pixbuf_get_has_alpha (pixbuf)) {
+		for (row = 0; row < height; row++) {
+			p = pixels + (row * row_stride);
+			for (column = 0; column < width; column++) {
+				r = *p++;
+				g = *p++;
+				b = *p++;
+				a = *p++;
+				
+				a_total += a;
+				r_total += r * a;
+				g_total += g * a;
+				b_total += b * a;
+			}
 		}
+		dividend = height * width * 0xFF;
+		a_total *= 0xFF;
+	} else {
+		for (row = 0; row < height; row++) {
+			p = pixels + (row * row_stride);
+			for (column = 0; column < width; column++) {
+				r = *p++;
+				g = *p++;
+				b = *p++;
+				
+				r_total += r;
+				g_total += g;
+				b_total += b;
+			}
+		}
+		dividend = height * width;
+		a_total = dividend * 0xFF;
 	}
 
-	a_total *= 0xFF;
-	dividend = height * width * 0xFF;
 	return ((a_total + dividend / 2) / dividend) << 24
 		| ((r_total + dividend / 2) / dividend) << 16
 		| ((g_total + dividend / 2) / dividend) << 8
@@ -377,7 +400,7 @@ eel_gdk_pixbuf_scale_down_to_fit (GdkPixbuf *pixbuf, int max_width, int max_heig
 	if (scale_factor >= 1.0) {
 		return gdk_pixbuf_copy (pixbuf);
 	} else {				
-		return gdk_pixbuf_scale_simple (pixbuf, scaled_width, scaled_height, GDK_INTERP_BILINEAR);	
+		return eel_gdk_pixbuf_scale_down (pixbuf, scaled_width, scaled_height);	
 	}
 }
 
@@ -425,7 +448,7 @@ eel_gdk_pixbuf_get_dimensions (const GdkPixbuf *pixbuf)
 }
 
 /**
- * eel_gdk_pixbuf_fill_rectangle:
+ * eel_gdk_pixbuf_fill_rectangle_with_color:
  * @pixbuf: Target pixbuf to fill into.
  * @area: Rectangle to fill.
  * @color: The color to use.
@@ -777,7 +800,7 @@ eel_gdk_pixbuf_draw_to_pixbuf_alpha (const GdkPixbuf *pixbuf,
 		
 		pixbuf = eel_gdk_pixbuf_new_from_pixbuf_sub_area ((GdkPixbuf *) pixbuf, area);
 	} else {
-		gdk_pixbuf_ref ((GdkPixbuf *) pixbuf);
+		g_object_ref (G_OBJECT (pixbuf));
 	}
 	
 	gdk_pixbuf_composite (pixbuf,
@@ -793,17 +816,17 @@ eel_gdk_pixbuf_draw_to_pixbuf_alpha (const GdkPixbuf *pixbuf,
 			      interpolation_mode,
 			      opacity);
 
-	gdk_pixbuf_unref ((GdkPixbuf *) pixbuf);
+	g_object_unref (G_OBJECT (pixbuf));
 }
 
 static void
-pixbuf_destroy_callback (guchar *pixels,
+pixbuf_destroy_callback (guchar  *pixels,
 			 gpointer callback_data)
 {
 	g_return_if_fail (pixels != NULL);
 	g_return_if_fail (callback_data != NULL);
 
-	gdk_pixbuf_unref ((GdkPixbuf *) callback_data);
+	g_object_unref (callback_data);
 }
 
 /**
@@ -1194,62 +1217,6 @@ eel_gdk_pixbuf_draw_to_drawable_tiled (const GdkPixbuf *pixbuf,
 			   &drawable_tile_data);
 }
 
-/**
- * eel_gdk_pixbuf_get_global_buffer:
- * @minimum_width: The minimum width for the requested buffer.
- * @minimum_height: The minimum height for the requested buffer.
- *
- * Return value: A GdkPixbuf that is at least as big as the passed in 
- *               dimensions.
- *
- * Access a global buffer for temporary GdkPixbuf operations.  
- * The returned buffer will be at least as big as the passed in 
- * dimensions.  The contents are not guaranteed to be anything at
- * anytime.  Also, it is not thread safe at all. */
-
-static GdkPixbuf *global_buffer = NULL;
-
-static void
-destroy_global_buffer (void)
-{
-	if (global_buffer != NULL) {
-		g_object_unref (global_buffer);
-		global_buffer = NULL;
-	}
-}
-
-GdkPixbuf *
-eel_gdk_pixbuf_get_global_buffer (int minimum_width,
-				  int minimum_height)
-{
-	static gboolean at_exit_deallocator_installed = FALSE;
-
-	g_return_val_if_fail (minimum_width > 0, NULL);
-	g_return_val_if_fail (minimum_height > 0, NULL);
-
-	if (global_buffer != NULL) {
-		if (gdk_pixbuf_get_width (global_buffer) >= minimum_width
-		    && gdk_pixbuf_get_height (global_buffer) >= minimum_height) {
-			return global_buffer;
-		}
-
-		destroy_global_buffer ();		
-	}
-
-	g_assert (global_buffer == NULL);
-
-	global_buffer = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, 
-					minimum_width, minimum_height);
-
-
-	if (at_exit_deallocator_installed == FALSE) {
-		at_exit_deallocator_installed = TRUE;
-		g_atexit (destroy_global_buffer);
-	}
-
-	return global_buffer;
-}
-
 /**
  * eel_gdk_pixbuf_intersect:
  * @pixbuf: A GdkPixbuf.
@@ -1290,34 +1257,38 @@ eel_gdk_pixbuf_intersect (const GdkPixbuf *pixbuf,
 }
 
 /**
- * eel_gdk_pixbuf_draw_layout
- * @pixbuf:    the pixbuf on which to draw the layout
- * @x:         the X position of the left of the layout (in pixels)
- * @y:         the Y position of the top of the layout (in pixels)
- * @color:     the color to use to draw the text, in "eel RGB" format
- * @layout:    a #PangoLayout
- *
- * Render a #PangoLayout onto a #GdkPixbuf.
- */
+ * eel_gdk_pixbuf_draw_layout_clipped:
+ * @pixbuf: the pixbuf to render to
+ * @region: text crop region
+ * @color: the color to render the text
+ * @layout: the text
+ * 
+ * This method renders the text in @layout into the
+ * @pixbuf with @color cropping it to @region.
+ **/
 void
-eel_gdk_pixbuf_draw_layout (GdkPixbuf *pixbuf,
-			    int x, int y,
-			    guint32 color,
-			    PangoLayout *layout)
+eel_gdk_pixbuf_draw_layout_clipped (GdkPixbuf   *pixbuf,
+				    ArtIRect     region,
+				    guint32      color,
+				    PangoLayout *layout)
 {
 	PangoRectangle ink_rect;
-	int pixbuf_height, pixbuf_width;
 	FT_Bitmap ink_bitmap;
 	GdkPixbuf *ink_pixbuf;
 	int bx, by;
+	int pixbuf_width, pixbuf_height;
 	const guint8 *ps;
 	guint8 *pd;
 	guint8 r, g, b;
+
+	if (region.x1 <= region.x0 || region.y1 <= region.y0) {
+		return;
+	}
 	
 	pango_layout_get_pixel_extents (layout, &ink_rect, NULL);
 
-	ink_rect.x += x;
-	ink_rect.y += y;
+	ink_rect.x += region.x0;
+	ink_rect.y += region.y0;
 
 	/* Do an intersect operation the long way.
 	 * Sad to write this all out like this.
@@ -1336,7 +1307,7 @@ eel_gdk_pixbuf_draw_layout (GdkPixbuf *pixbuf,
 	pixbuf_width = gdk_pixbuf_get_width (pixbuf);
 	if (ink_rect.x + ink_rect.width > pixbuf_width) {
 		ink_rect.width = pixbuf_width - ink_rect.x;
-		if (ink_rect.width < 0) {
+		if (ink_rect.width <= 0) {
 			return;
 		}
 	}
@@ -1344,7 +1315,7 @@ eel_gdk_pixbuf_draw_layout (GdkPixbuf *pixbuf,
 	pixbuf_height = gdk_pixbuf_get_height (pixbuf);
 	if (ink_rect.y + ink_rect.height > pixbuf_height) {
 		ink_rect.height = pixbuf_height - ink_rect.y;
-		if (ink_rect.height < 0) {
+		if (ink_rect.height <= 0) {
 			return;
 		}
 	}
@@ -1356,7 +1327,9 @@ eel_gdk_pixbuf_draw_layout (GdkPixbuf *pixbuf,
 	ink_bitmap.num_grays = 0x100;
 	ink_bitmap.pixel_mode = ft_pixel_mode_grays;
 
-	pango_ft2_render_layout (&ink_bitmap, layout, x - ink_rect.x, y - ink_rect.y);
+	pango_ft2_render_layout (&ink_bitmap, layout,
+				 region.x0 - ink_rect.x,
+				 region.y0 - ink_rect.y);
 
 	ink_pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
 				     ink_bitmap.width, ink_bitmap.rows);
@@ -1381,13 +1354,163 @@ eel_gdk_pixbuf_draw_layout (GdkPixbuf *pixbuf,
 
 	gdk_pixbuf_composite (ink_pixbuf, pixbuf,
 			      ink_rect.x, ink_rect.y,
-			      ink_bitmap.width, ink_bitmap.rows,
+			      MIN (region.x1 - region.x0, ink_bitmap.width),
+			      MIN (region.y1 - region.y0, ink_bitmap.rows),
 			      ink_rect.x, ink_rect.y,
-			      1, 1, GDK_INTERP_BILINEAR, 0xFF);
+			      1.0, 1.0, GDK_INTERP_BILINEAR, 0xFF);
 
 	g_object_unref (ink_pixbuf);
 }
 
+/**
+ * eel_gdk_pixbuf_draw_layout
+ * @pixbuf:    the pixbuf on which to draw the layout
+ * @x:         the X position of the left of the layout (in pixels)
+ * @y:         the Y position of the top of the layout (in pixels)
+ * @color:     the color to use to draw the text, in "eel RGB" format
+ * @layout:    a #PangoLayout
+ *
+ * Render a #PangoLayout onto a #GdkPixbuf.
+ */
+void
+eel_gdk_pixbuf_draw_layout (GdkPixbuf *pixbuf,
+			    int x, int y,
+			    guint32 color,
+			    PangoLayout *layout)
+{
+	ArtIRect region;
+
+	region.x0 = x;
+	region.y0 = y;
+	region.x1 = region.y1 = G_MAXINT;
+
+	eel_gdk_pixbuf_draw_layout_clipped (pixbuf, region, color, layout);
+}
+
+GdkPixbuf *
+eel_gdk_pixbuf_scale_down (GdkPixbuf *pixbuf,
+			   int dest_width,
+			   int dest_height)
+{
+	int source_width, source_height;
+	int s_x1, s_y1, s_x2, s_y2;
+	int s_xfrac, s_yfrac;
+	int dx, dx_frac, dy, dy_frac;
+	div_t ddx, ddy;
+	int x, y;
+	int r, g, b, a;
+	int n_pixels;
+	gboolean has_alpha;
+	guchar *dest, *src, *xsrc, *src_pixels;
+	GdkPixbuf *dest_pixbuf;
+	int pixel_stride;
+	int source_rowstride, dest_rowstride;
+
+	if (dest_width == 0 || dest_height == 0) {
+		return NULL;
+	}
+	
+	source_width = gdk_pixbuf_get_width (pixbuf);
+	source_height = gdk_pixbuf_get_height (pixbuf);
+
+	g_assert (source_width >= dest_width);
+	g_assert (source_height >= dest_height);
+
+	ddx = div (source_width, dest_width);
+	dx = ddx.quot;
+	dx_frac = ddx.rem;
+	
+	ddy = div (source_height, dest_height);
+	dy = ddy.quot;
+	dy_frac = ddy.rem;
+
+	has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
+	source_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
+	src_pixels = gdk_pixbuf_get_pixels (pixbuf);
+
+	dest_pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, has_alpha, 8,
+				      dest_width, dest_height);
+	dest = gdk_pixbuf_get_pixels (dest_pixbuf);
+	dest_rowstride = gdk_pixbuf_get_rowstride (dest_pixbuf);
+
+	pixel_stride = (has_alpha)?4:3;
+	
+	s_y1 = 0;
+	s_yfrac = -dest_height/2;
+	while (s_y1 < source_height) {
+		s_y2 = s_y1 + dy;
+		s_yfrac += dy_frac;
+		if (s_yfrac > 0) {
+			s_y2++;
+			s_yfrac -= dest_height;
+		}
+
+		s_x1 = 0;
+		s_xfrac = -dest_width/2;
+		while (s_x1 < source_width) {
+			s_x2 = s_x1 + dx;
+			s_xfrac += dx_frac;
+			if (s_xfrac > 0) {
+				s_x2++;
+				s_xfrac -= dest_width;
+			}
+
+			/* Average block of [x1,x2[ x [y1,y2[ and store in dest */
+			r = g = b = a = 0;
+			n_pixels = 0;
+
+			src = src_pixels + s_y1 * source_rowstride + s_x1 * pixel_stride;
+			for (y = s_y1; y < s_y2; y++) {
+				xsrc = src;
+				if (has_alpha) {
+					for (x = 0; x < s_x2-s_x1; x++) {
+						n_pixels++;
+						
+						r += xsrc[3] * xsrc[0];
+						g += xsrc[3] * xsrc[1];
+						b += xsrc[3] * xsrc[2];
+						a += xsrc[3];
+						xsrc += 4;
+					}
+				} else {
+					for (x = 0; x < s_x2-s_x1; x++) {
+						n_pixels++;
+						r += *xsrc++;
+						g += *xsrc++;
+						b += *xsrc++;
+					}
+				}
+				src += source_rowstride;
+			}
+			
+			if (has_alpha) {
+				if (a != 0) {
+					*dest++ = r / a;
+					*dest++ = g / a;
+					*dest++ = b / a;
+					*dest++ = a / n_pixels;
+				} else {
+					*dest++ = 0;
+					*dest++ = 0;
+					*dest++ = 0;
+					*dest++ = 0;
+				}
+			} else {
+				*dest++ = r / n_pixels;
+				*dest++ = g / n_pixels;
+				*dest++ = b / n_pixels;
+			}
+			
+			s_x1 = s_x2;
+		}
+		s_y1 = s_y2;
+		dest += dest_rowstride - dest_width * pixel_stride;
+	}
+	
+	return dest_pixbuf;
+}
+
+
 #if !defined (EEL_OMIT_SELF_CHECK)
 
 static char *
@@ -1535,3 +1658,4 @@ eel_self_check_gdk_pixbuf_extensions (void)
 }
 
 #endif /* !EEL_OMIT_SELF_CHECK */
+
